#ifndef _INPUT_H
#define _INPUT_H 1

#include <vector>
#include <string>

class InputProto;

class Input {
public:
	Input(const std::string &stream_id, const std::string &url);

	// Serialization/deserialization.
	Input(const InputProto &serialized);
	InputProto serialize() const;

	// Connect to the given URL and start streaming.
	void run();

	// Stops the streaming, but lets the file descriptor stay open.
	void stop();

	std::string get_url() const { return url; }

private:
	// Recovers the this pointer and calls do_work().
	static void *do_work_thunk(void *arg);

	// Actually does the download.
	void do_work();
	
	// Open a socket that connects to the given host and port. Does DNS resolving.
	int lookup_and_connect(const std::string &host, const std::string &port);

	// Parses a HTTP response. Returns false if it not a 200.
	bool parse_response(const std::string &response);

	// Stores the given data, looks for Metacube blocks (skipping data if needed),
	// and calls process_block() for each one.
	void process_data(char *ptr, size_t bytes);

	// Drops <num_bytes> bytes from the head of <pending_data>,
	// and outputs a warning.
	void drop_pending_data(size_t num_bytes);

	enum State {
		NOT_CONNECTED,
		SENDING_REQUEST,
		RECEIVING_HEADER,
		RECEIVING_DATA,
		CLOSING_SOCKET,  // Due to error.
	};
	State state;

	std::string stream_id;

	// The URL and its parsed components.
	std::string url;
	std::string host, port, path;

	// The HTTP request, with headers and all.
	// Only relevant for SENDING_REQUEST.
	std::string request;

	// How many bytes we've sent of the request so far.
	// Only relevant for SENDING_REQUEST.
	size_t request_bytes_sent;

	// The HTTP response we've received so far. Only relevant for RECEIVING_HEADER.
	std::string response;

	// The HTTP respones headers we want to give clients for this input.
	std::string http_header;

	// Data we have received but not fully processed yet.
	std::vector<char> pending_data;

	// If <pending_data> starts with a Metacube header,
	// this is true.
	bool has_metacube_header;

	// The socket we are downloading on (or -1).
	int sock;	

	// Handle to the thread that actually does the download.
	pthread_t worker_thread;

	// Whether we should stop or not.
	volatile bool should_stop;
};

#endif  // !defined(_INPUT_H)
